"""R2X Sienna system exporter."""

# System packages
import json
import os
from functools import partial
from operator import itemgetter
from typing import Any
from urllib.request import urlopen

# Third-party packages
from loguru import logger

# Local imports
from r2x.config_models import ReEDSConfig, SiennaConfig
from r2x.config_utils import get_year
from r2x.exporter.handler import BaseExporter, get_export_records
from r2x.exporter.utils import (
    apply_default_value,
    apply_flatten_key,
    apply_pint_deconstruction,
    apply_property_map,
    apply_unnest_key,
)
from r2x.models import (
    ACBranch,
    Bus,
    DCBranch,
    Generator,
    HydroPumpedStorage,
    PowerLoad,
    Reserve,
    ReserveMap,
    Storage,
)
from r2x.models.branch import Transformer2W
from r2x.utils import haskey

PSY_URL = "https://raw.githubusercontent.com/NREL-Sienna/PowerSystems.jl/refs/heads/main/"
TABLE_DATA_SPEC = "src/descriptors/power_system_inputs.json"


def get_psy_fields() -> dict[str, Any]:
    """Get PSY JSON schema."""
    request = urlopen(PSY_URL + TABLE_DATA_SPEC)
    descriptor = json.load(request)
    return descriptor


class SiennaExporter(BaseExporter):
    """Sienna exporter class.

    This is the class that export the infrasys system to csv's required to read
    in PowerSystems.jl. The class contains a method per file that needs to get
    created and some of the methods manipulate the data to be compliant wth
    PowerSystems.jl. The order is important for some steps, however, most
    of them could be run sequentially.

    The method `run` is the responsible of running all the exporter.

    Notes
    -----
    All the methods defined an `output_columns` which defines the fields from
    the `Component` that we want to export.

    All methods call `export_component_to_csv` or `export_dict_to_csv` that
    is defined on the `BaseExporter` class.


    See Also
    --------
    BaseExporter.export_component_to_csv
    BaseExporter.export_dict_to_csv
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert self.config.output_config
        assert self.config.input_config
        self.output_data = {}
        if not isinstance(self.config.output_config, SiennaConfig):
            msg = (
                f"Output config is of type {type(self.config.output_config)}. "
                "It should be type of `PlexosConfig`."
            )
            raise TypeError(msg)
        if isinstance(self.config.input_config, ReEDSConfig) and isinstance(
            self.config.input_config.solve_year, list
        ):
            msg = "Multiple solve years are not supported yet."
            raise NotImplementedError(msg)

        self.output_config = self.config.input_config.to_class(SiennaConfig, self.config.output_config)
        self.property_map = self.output_config.defaults.get("sienna_property_map", {})
        self.unit_map = self.output_config.defaults.get("sienna_unit_map", {})
        self.output_fields = self.output_config.defaults["table_data"]
        self.year = get_year(self.output_config)
        assert self.year is not None

    def run(self, *args, path=None, **kwargs) -> "SiennaExporter":
        """Run sienna exporter workflow.


        Notes
        -----
        All methods call `export_component_to_csv` or `export_dict_to_csv` that
        is defined on the `BaseExporter` class.

        If performance is a constraint, we could parallelize the methods.
        """
        logger.info("Starting {}", self.__class__.__name__)

        self.process_bus_data()
        self.process_load_data()
        self.process_branch_data()
        self.process_dc_branch_data()
        self.process_gen_data()
        self.process_reserves_data()
        self.process_storage_data()
        self.export_data()
        self.create_timeseries_pointers()
        return self

    def process_bus_data(self, fname: str = "bus.csv") -> None:
        """Create bus.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "bus_id",
            "name",
            "area",
            "zone",
            "base_voltage",
            "bus_type",
        ]

        records = [
            component.model_dump(exclude_none=True, mode="python", serialize_as_any=True)
            for component in self.system.get_components(Bus)
        ]

        key_mapping = {"number": "bus_id", "load_zone": "zone"}
        export_records = get_export_records(
            records,
            partial(apply_property_map, property_map=self.property_map | key_mapping),
            partial(apply_pint_deconstruction, unit_map=self.unit_map),
            partial(apply_unnest_key, key_map={"zone": "name", "area": "name"}),
        )
        self.system._export_dict_to_csv(
            export_records,
            fpath=self.output_folder / fname,
            fields=output_fields,
            restval="NA",
        )

    def process_load_data(self, fname: str = "load.csv") -> None:
        """Create load.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "bus_id",
            "name",
            "available",
            "active_power",
            "reactive_power",
            "max_active_power",
            "max_reeactive_power",
        ]
        records = [
            component.model_dump(exclude_none=True, mode="python", serialize_as_any=True)
            for component in self.system.get_components(PowerLoad)
        ]
        key_mapping = {
            "bus": "bus_id",
        }
        export_records = get_export_records(
            records,
            partial(apply_property_map, property_map=self.property_map | key_mapping),
            partial(apply_pint_deconstruction, unit_map=self.unit_map),
            partial(apply_unnest_key, key_map={"bus_id": "number"}),
        )
        self.system._export_dict_to_csv(
            export_records,
            fpath=self.output_folder / fname,
            fields=output_fields,
            restval="0.0",
        )
        logger.info(f"File {fname} created.")

    def process_branch_data(self, fname: str = "branch.csv") -> None:
        """Create branch.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "connection_points_from",
            "connection_points_to",
            "r",
            "x",
            "primary_shunt",
            "rate",
            "rating_up",
            "rating_down",
            "tap",
            "is_transformer",
            "ext",
        ]

        key_mapping = {
            "from_bus": "connection_points_from",
            "to_bus": "connection_points_to",
            "class_type": "branch_type",
            "rating": "rate",
            "b": "primary_shunt",
        }

        records = [
            component.model_dump(exclude_none=True, mode="python", serialize_as_any=True)
            for component in self.system.get_components(ACBranch)
        ]
        export_records = get_export_records(
            records,
            partial(apply_property_map, property_map=self.property_map | key_mapping),
            partial(apply_pint_deconstruction, unit_map=self.unit_map),
            partial(
                apply_unnest_key,
                key_map={"connection_points_from": "number", "connection_points_to": "number"},
            ),
            partial(
                lambda component, key, func: component.update({key: func(component)}) or component,
                key="is_transformer",
                func=lambda d: True if d["branch_type"] == Transformer2W.__name__ else False,
            ),
            partial(apply_default_value, default_value_map={"tap": 1.0}),
        )
        self.system._export_dict_to_csv(
            export_records,
            fpath=self.output_folder / fname,
            fields=output_fields,
            restval="NA",
        )
        logger.info(f"File {fname} created.")

    def process_dc_branch_data(self, fname="dc_branch.csv") -> None:
        """Create dc_branch.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "connection_points_from",
            "connection_points_to",
            "rate",
            "loss",
        ]

        self.system.export_component_to_csv(
            DCBranch,
            fpath=self.output_folder / fname,
            fields=output_fields,
            unnest_key="number",
            key_mapping={
                "from_bus": "connection_points_from",
                "to_bus": "connection_points_to",
                "class_type": "branch_type",
                "rating_up": "rate",
            },
            restval="NA",
        )
        logger.info(f"File {fname} created.")
        return

    def process_gen_data(self, fname="gen.csv"):
        """Create gen.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        # reactive power cant be export

        key_mapping = {
            "bus": "bus_id",
            "prime_mover_type": "unit_type",
        }

        records = [
            component.model_dump(exclude_none=True, mode="python", serialize_as_any=True)
            for component in self.system.get_components(Generator)
        ]
        export_records = get_export_records(
            records,
            partial(apply_operation_table_data),
            partial(apply_flatten_key, keys_to_flatten={"active_power_limits"}),
            partial(apply_property_map, property_map=self.property_map | key_mapping),
            partial(apply_pint_deconstruction, unit_map=self.unit_map),
            partial(apply_unnest_key, key_map={"bus_id": "number"}),
            partial(
                apply_default_value,
                default_value_map={"fuel_price": 0.0, "power_factor": 1.0, "startup_cost": 0.0},
            ),
        )

        sorted_records = sorted(export_records, key=itemgetter("name"), reverse=True)
        self.system._export_dict_to_csv(
            sorted_records,
            fpath=self.output_folder / fname,
            fields=self.output_fields["generator"],
            restval="NA",
        )
        logger.info(f"File {fname} created.")

    def process_reserves_data(self, fname="reserves.csv") -> None:
        """Create reserve.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "requirement",
            "timeframe",
            "eligible_region",
            "direction",
            "contributing_devices",
            "eligible_device_categories",
            "eligible_device_subcategories",
        ]

        # ReserveMap holds the eligible devices in the mapping.
        reserve_map_list: list[dict[str, Any]] = list(
            map(lambda component: component["mapping"], self.system.to_records(ReserveMap))
        )

        if not reserve_map_list:
            msg = "Reserve map class not found on the system. Skipping reserve contributing devices file."
            logger.warning(msg)
            return

        if len(reserve_map_list) > 1:
            logger.warning("We do not support multiple reserve maps per system")
            return
        reserve_map: dict = reserve_map_list[0]

        reserves: list[dict[str, Any]] = list(self.system.to_records(Reserve))

        output_data = []
        for reserve in reserves:
            # NOTE: Only export reserves that exists on the reserve map
            if reserve["name"] in reserve_map:
                output_dict = reserve
                output_dict["direction"] = reserve["direction"].name
                output_dict["eligible_device_categories"] = "(Generator,Storage)"
                contributing_devices = reserve_map.get(reserve["name"])
                output_dict["contributing_devices"] = str(tuple(contributing_devices)).replace(  # type: ignore
                    "'", ""
                )  # Sienna expects a tuple for this field that looks like "(val1,val2,val3)"
                output_data.append(output_dict)

        key_mapping = {"region": "eligible_region", "max_requirement": "requirement"} | self.property_map

        export_records = get_export_records(
            output_data,
            partial(apply_property_map, property_map=self.property_map | key_mapping),
            partial(apply_pint_deconstruction, unit_map=self.unit_map),
            partial(apply_unnest_key, key_map={"eligible_region": "name"}),
        )
        self.system._export_dict_to_csv(
            export_records,
            fpath=self.output_folder / fname,
            fields=output_fields,
            restval="NA",
        )
        logger.info(f"File {fname} created.")
        return

    def process_storage_data(self, fname="storage.csv") -> None:
        """Create storage.csv file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        output_fields = [
            "name",
            "position",
            "available",
            "generator_name",
            "bus_id",
            "active_power",
            "rating",
            "input_efficiency",
            "output_efficiency",
            "storage_capacity",
            "min_storage_capacity",
            "max_storage_capacity",
            "input_active_power_limit_min",
            "output_active_power_limit_max",
            "output_active_power_limit_min",
            "unit_type",
        ]

        storage_list = get_export_records(
            list(
                self.system.to_records(
                    Generator, filter_func=lambda x: isinstance(x, Storage | HydroPumpedStorage)
                )
            ),
            partial(apply_property_map, property_map=self.property_map),
            partial(apply_flatten_key, keys_to_flatten={"active_power_limits"}),
            partial(apply_pint_deconstruction, unit_map=self.unit_map),
            # partial(apply_valid_properties, valid_properties=output_fields),
        )

        if not storage_list:
            logger.warning("No storage devices found")
            return

        output_data = []
        for storage in storage_list:
            output_dict = storage
            output_dict["generator_name"] = storage["name"]
            output_dict["input_active_power_limit_max"] = output_dict["active_power_limits_max"]
            output_dict["output_active_power_limit_max"] = output_dict["active_power_limits_max"]
            # NOTE: If we need to change this in the future, we could probably
            # use the function max to check if the component has the field.
            output_dict["input_active_power_limit_min"] = 0  # output_dict["active_power"]
            output_dict["output_active_power_limit_min"] = 0  # output_dict["active_power"]
            output_dict["active_power"] = output_dict["active_power"]
            output_dict["bus_id"] = (
                getattr(self.system.get_component_by_label(output_dict["bus"]), "number", None)
                if output_dict["bus"]
                else None
            )
            output_dict["rating"] = output_dict["rating"]

            # NOTE: For pumped hydro storage we create a head and a tail
            # representation that keeps track of the upper and down reservoir
            # state of charge.
            # if storage["class_type"] == "HydroPumpedStorage":
            original_name = output_dict["name"]
            output_dict["name"] = original_name + "_head"
            output_dict["position"] = "head"
            output_data.append(output_dict)
            tail_copy = output_dict.copy()
            tail_copy["name"] = original_name + "_tail"
            output_dict["position"] = "tail"
            output_data.append(tail_copy)

        key_mapping = self.property_map
        self.system._export_dict_to_csv(
            output_data,
            fpath=self.output_folder / fname,
            fields=output_fields,
            key_mapping=key_mapping,
            unnest_key="number",
            restval="NA",
        )

        logger.info("File storage.csv created.")

    def create_timeseries_pointers(self) -> None:
        """Create timeseries_pointers.json file.

        Parameters
        ----------
        fname : str
            Name of the file to be created
        """
        ts_pointers_list = []

        for component_type, time_series in self.time_series_objects.items():
            csv_fpath = self.ts_directory / (f"{component_type}_{self.config.name}_{self.year}.csv")
            for i in range(len(time_series)):
                component_name = self.time_series_name_by_type[component_type][i]
                ts_instance = time_series[i]
                resolution = ts_instance.resolution.seconds
                variable_name = self.property_map.get(ts_instance.variable_name, ts_instance.variable_name)
                # Set to None if the variable is not found in the property map
                ts_pointers = {
                    "category": component_type.split("_", maxsplit=1)[0],  # Component_name is the first
                    "component_name": component_name,
                    "data_file": str(csv_fpath),
                    "normalization_factor": None,
                    "resolution": resolution,
                    "name": variable_name,
                    "scaling_factor_multiplier_module": "PowerSystems",
                    "scaling_factor_multiplier": None,
                }
                ts_pointers_list.append(ts_pointers)

        # Sort the list to make it easier to visually diff
        ts_pointers_list.sort(key=lambda x: x["component_name"])

        with open(os.path.join(self.output_folder, "timeseries_pointers.json"), mode="w") as f:
            json.dump(ts_pointers_list, f, indent=4)

        logger.info("File timeseries_pointers.json created.")
        return

    def export_data(self) -> None:
        """Export csv data to specified folder from output_data attribute."""
        logger.debug("Saving Sienna data and timeseries files.")

        # First export all time series objects
        self.time_series_to_csv(config=self.config, system=self.system, reference_year=self.year)
        logger.info("Saving time series data.")


def apply_operation_table_data(
    component: dict[str, Any],
) -> dict[str, Any]:
    """Process and apply operation cost data for `PSY.power_system_table_data.jl`.

    This function extracts operation cost data from a component dictionary and adds
    various cost-related fields to it. It handles different types of cost data,
    including variable costs, fuel costs, heat rates, and cost/fuel curves.

    Parameters
    ----------
    component : dict[str, Any]
        A dictionary containing component data, potentially including operation cost information.

    Returns
    -------
    dict[str, Any]
        The input component dictionary, potentially modified with additional operation cost fields.

    Notes
    -----
    This function modifies the input dictionary in-place and returns it.

    The function processes the following types of operation costs:
    - Variable costs (vom_cost)
    - Fuel costs
    - Heat rate coefficients (a0, a1, a2)
    - Cost curves
    - Fuel curves

    Raises
    ------
    AssertionError
        If the fuel_cost is None when present in the variable cost data.
    NotImplementedError
        If an unsupported variable curve type is encountered.

    Examples
    --------
    >>> component = {
    ...     "operation_cost": {
    ...         "variable": {
    ...             "vom_cost": {"function_data": {"proportional_term": 10}},
    ...             "fuel_cost": 0.05,
    ...             "value_curve": {
    ...                 "function_data": {
    ...                     "constant_term": 100,
    ...                     "proportional_term": 20,
    ...                     "quadratic_term": 0.5,
    ...                     "points": [(0, 0), (50, 1000), (100, 2500)],
    ...                 }
    ...             },
    ...         },
    ...         "variable_type": "CostCurve",
    ...     }
    ... }
    >>> updated_component = apply_operation_table_data(component)
    >>> updated_component["variable_cost"]
    10
    >>> updated_component["fuel_price"]
    50.0
    >>> updated_component["heat_rate_a0"]
    100
    >>> updated_component["output_point_1"]
    50
    >>> updated_component["cost_point_1"]
    1000
    """
    if not component.get("operation_cost", False):
        return component

    operation_cost = component["operation_cost"]

    if not (variable := operation_cost.get("variable")):
        return component

    if haskey(variable, ["vom_cost", "function_data"]):
        component["variable_cost"] = variable["vom_cost"]["function_data"].proportional_term

    if "fuel_cost" in variable.keys():
        assert variable["fuel_cost"] is not None
        # Note: We multiply the fuel price by 1000 to offset the division
        # done by Sienna when it parses .csv files
        component["fuel_price"] = variable["fuel_cost"] * 1000
    if haskey(variable, ["value_curve", "function_data"]):
        function_data = variable["value_curve"]["function_data"]
        function_data_fields = function_data.model_fields_set
        if "constant_term" in function_data_fields:
            component["heat_rate_a0"] = function_data.constant_term
        if "proportional_term" in function_data_fields:
            component["heat_rate_a1"] = function_data.proportional_term
        if "quadratic_term" in function_data_fields:
            component["heat_rate_a2"] = function_data.quadratic_term
        if "points" in function_data_fields:
            component = _variable_type_parsing(component, operation_cost)
    return component


def _variable_type_parsing(component: dict, cost_dict: dict[str, Any]) -> dict[str, Any]:
    variable_curve = cost_dict["variable"]
    x_y_coords = variable_curve["value_curve"]["function_data"].points
    match cost_dict["variable_type"]:
        case "CostCurve":
            for i, (x_coord, y_coord) in enumerate(x_y_coords):
                output_point_col = f"output_point_{i}"
                component[output_point_col] = x_coord

                cost_point_col = f"cost_point_{i}"
                component[cost_point_col] = y_coord
        case "FuelCurve":
            for i, (x_coord, y_coord) in enumerate(x_y_coords):
                output_point_col = f"output_point_{i}"
                component[output_point_col] = x_coord

                heat_rate_col = "heat_rate_avg_0" if i == 0 else f"heat_rate_incr_{i}"
                component[heat_rate_col] = y_coord
        case _:
            msg = f"Type {cost_dict['variable_type']} variable curve not supported"
            raise NotImplementedError(msg)
    return component



from infrasys.cost_curves import FuelCurve, UnitSystem
from infrasys.value_curves import LinearCurve
from r2x.models.costs import (
    HydroGenerationCost,
    OperationalCost,
    RenewableGenerationCost,
    StorageCost,
    ThermalGenerationCost,
)


def test_properties():
    cost = OperationalCost()
    assert isinstance(cost, OperationalCost)


def test_computed_fields():
    variable = FuelCurve(value_curve=LinearCurve(0), power_units=UnitSystem.NATURAL_UNITS)
    cost = ThermalGenerationCost(variable=variable)

    assert isinstance(cost, OperationalCost)
    assert isinstance(cost, ThermalGenerationCost)
    assert cost.variable_type == "FuelCurve"
    assert cost.value_curve_type == "InputOutputCurve"

    cost = ThermalGenerationCost()
    assert isinstance(cost, OperationalCost)
    assert isinstance(cost, ThermalGenerationCost)
    assert cost.variable_type is None
    assert cost.value_curve_type is None


def test_default_fields():
    cost = ThermalGenerationCost()
    assert isinstance(cost, OperationalCost)
    assert isinstance(cost, ThermalGenerationCost)
    assert cost.fixed == 0.0
    assert cost.shut_down == 0.0
    assert cost.start_up == 0.0

    cost = HydroGenerationCost()
    assert isinstance(cost, OperationalCost)
    assert isinstance(cost, HydroGenerationCost)
    assert cost.fixed == 0.0

    cost = RenewableGenerationCost()
    assert isinstance(cost, OperationalCost)
    assert isinstance(cost, RenewableGenerationCost)
    assert cost.curtailment_cost is None

    cost = StorageCost()
    assert isinstance(cost, OperationalCost)
    assert isinstance(cost, StorageCost)
    assert cost.energy_surplus_cost == 0.0
    assert cost.fixed == 0.0
    assert cost.shut_down == 0.0
    assert cost.start_up == 0.0



import pytest
from datetime import datetime, timedelta
import polars as pl
from r2x.parser.plexos_utils import DATAFILE_COLUMNS, get_column_enum, time_slice_handler


def test_get_column_enum():
    """Test multiple cases for get_column_enum function."""
    columns = ["year", "random_column", "random_column_2"]
    assert get_column_enum(columns) == DATAFILE_COLUMNS.Y

    columns = ["name", "value"]
    assert get_column_enum(columns) == DATAFILE_COLUMNS.NV

    # Case 2: Exact match for TS_YMDPV
    columns = ["year", "month", "day", "period", "value"]
    assert get_column_enum(columns) == DATAFILE_COLUMNS.TS_YMDPV

    # Case 3: Subset match for TS_YM (with an extra column)
    columns = ["year", "month", "extra"]
    assert get_column_enum(columns) == DATAFILE_COLUMNS.TS_YM

    # Case 4: No match (completely unrelated columns)
    columns = ["random", "columns"]
    assert get_column_enum(columns) is None

    # Case 5: Partial match for NV (extra column in input)
    columns = ["name", "value", "extra"]
    assert get_column_enum(columns) == DATAFILE_COLUMNS.NV

    # Case 6: Exact match for TS_NMDH (large set of columns)
    columns = [
        "name",
        "month",
        "day",
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
        "11",
        "12",
        "13",
        "14",
        "15",
        "16",
        "17",
        "18",
        "19",
        "20",
        "21",
        "22",
        "23",
        "24",
    ]
    assert get_column_enum(columns) == DATAFILE_COLUMNS.TS_NMDH


def test_time_slice_handler():
    records = [{"pattern": "M1-2", "value": 200}, {"pattern": "M3-12", "value": 100}]

    year = 2012
    hourly_time_index = pl.datetime_range(
        datetime(year, 1, 1), datetime(year + 1, 1, 1), interval="1h", eager=True, closed="left"
    ).to_frame("datetime")
    result_polars = time_slice_handler(records, hourly_time_index)

    assert all(result_polars[:100] == 200)
    assert all(result_polars[-100:] == 100)

    start = datetime(year, 1, 1)
    end = datetime(year + 1, 1, 1)
    delta = timedelta(hours=1)
    datetime_index = tuple(start + i * delta for i in range((end - start) // delta))
    result_datetime = time_slice_handler(records, datetime_index)
    assert all(result_datetime == result_polars)


def test_time_slice_handler_raises():
    year = 2020
    start = datetime(year, 1, 1)
    end = datetime(year + 1, 1, 1)
    delta = timedelta(hours=1)
    datetime_index = tuple(start + i * delta for i in range((end - start) // delta))
    records = [{"pattern": "M1-2", "value": 200}, {"pattern": "M3-12", "value": 100}, [1, 2]]
    with pytest.raises(TypeError):
        _ = time_slice_handler(records, datetime_index)

    records = [{"pattern": "H1-2", "value": 200}, {"pattern": "M3-12", "value": 100}]
    with pytest.raises(NotImplementedError):
        _ = time_slice_handler(records, datetime_index)



from datetime import datetime, timedelta
from pathlib import Path

import pytest
from infrasys import SingleTimeSeries

from r2x.api import System
from r2x.config_scenario import Scenario
from r2x.exporter import SiennaExporter
from r2x.models import ACBus, RenewableDispatch


@pytest.fixture
def scenario_instance(data_folder, tmp_folder):
    return Scenario.from_kwargs(
        name="TestScenario",
        run_folder=data_folder,
        output_folder=tmp_folder,
        input_model="infrasys",
        output_model="sienna",
        model_year=2010,
        reference_year=2010,
    )


@pytest.fixture
def sienna_exporter(scenario_instance, infrasys_test_system, tmp_folder):
    return SiennaExporter(config=scenario_instance, system=infrasys_test_system, output_folder=tmp_folder)


def test_export_data_files(scenario_instance, tmp_folder, sienna_exporter):
    system = sienna_exporter.system
    default_ts_directory = "Data"
    sienna_exporter.time_series_to_csv(
        scenario_instance,
        system,
        output_folder=tmp_folder,
        time_series_folder=default_ts_directory,
        time_series_fname=getattr(scenario_instance, "time_series_fname", None),
    )

    assert scenario_instance.output_folder
    assert (scenario_instance.output_folder / default_ts_directory).exists()
    time_series_fpath = scenario_instance.output_folder / default_ts_directory
    assert sum(1 for _ in Path(time_series_fpath).iterdir() if _.is_file()) == 3  # Count files


def test_time_series_to_csv_with_user_attributes(scenario_instance, tmp_folder):
    system = System()
    bus = ACBus(name="TestBus", number=1)
    gen1 = RenewableDispatch(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = RenewableDispatch(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)

    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = range(length)
    variable_name = "active_power"
    ts = SingleTimeSeries.from_time_array(data, variable_name, time_array)
    system.add_time_series(ts, gen1, model_year=2020)

    default_ts_directory = "Data"
    _ = SiennaExporter.time_series_to_csv(
        scenario_instance, system, output_folder=tmp_folder, model_year=2025
    )
    assert scenario_instance.output_folder
    assert (scenario_instance.output_folder / default_ts_directory).exists()
    time_series_fpath = scenario_instance.output_folder / default_ts_directory
    assert sum(1 for _ in Path(time_series_fpath).iterdir() if _.is_file()) == 0  # Count files


def test_time_series_to_csv_with_multiple_lengths(scenario_instance):
    system = System()
    bus = ACBus(name="TestBus", number=1)
    gen1 = RenewableDispatch(name="gen1", active_power=1.0, rating=1.0, bus=bus, available=True)
    gen2 = RenewableDispatch(name="gen2", active_power=1.0, rating=1.0, bus=bus, available=True)
    system.add_components(bus, gen1, gen2)

    # Length 10 time series
    length = 10
    initial_time = datetime(year=2020, month=1, day=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = range(length)
    variable_name = "active_power"
    ts = SingleTimeSeries.from_time_array(data, variable_name, time_array)
    system.add_time_series(ts, gen1)

    length = 20
    initial_time = datetime(year=2020, month=1, day=1)
    time_array = [initial_time + timedelta(hours=i) for i in range(length)]
    data = range(length)
    variable_name = "active_power"
    ts2 = SingleTimeSeries.from_time_array(data, variable_name, time_array)
    system.add_time_series(ts2, gen2)
    with pytest.raises(NotImplementedError):
        _ = SiennaExporter.time_series_to_csv(scenario_instance, system)



"""Helper functions for the exporters."""

from typing import Any
from collections.abc import Callable
import copy
from functools import wraps
from r2x.enums import ReserveType, ReserveDirection
from r2x.exceptions import FieldRemovalError
import pint
from infrasys.base_quantity import BaseQuantity


def get_reserve_type(
    reserve_type: ReserveType, reserve_direction: ReserveDirection, reserve_types: dict[str, dict[str, str]]
) -> str:
    """Return the reserve type from a mapping.

    If not found, return default reserve type
    """
    for key, value in reserve_types.items():
        if value["type"] == reserve_type and value["direction"] == reserve_direction:
            return key
    return get_reserve_type(
        ReserveType[reserve_types["default"]["type"]],
        ReserveDirection[reserve_types["default"]["direction"]],
        reserve_types,
    )


def required_fields(*fields: str | list[str] | set):
    """Specify required fields for the transformation."""

    def decorator(
        func: Callable,
    ) -> Callable:
        @wraps(func)
        def wrapper(component_data, *args, **kwargs):
            original_data = copy.deepcopy(component_data)
            result = func(component_data)
            removed_fields = set(original_data.keys()) - set(result.keys())
            if removed_fields & set(fields):
                removed_required = removed_fields & set(fields)
                raise FieldRemovalError(
                    f"Transformation {func.__name__} removed required fields: {removed_required}"
                )
            return result

        return wrapper

    return decorator


def compose(
    *functions: Callable[[dict[str, Any]], dict[str, Any]],
) -> Callable[[dict[str, Any]], dict[str, Any]]:
    """
    Compose multiple functions, applying them sequentially to the input data.

    Parameters
    ----------
    functions : Callable[[Dict[str, Any]], Dict[str, Any]]
        Functions that transform a dictionary.

    Returns
    -------
    Callable[[Dict[str, Any]], Dict[str, Any]]
        A function that sequentially applies the provided functions to a dictionary.
    """

    def apply_all(data: dict[str, Any]) -> dict[str, Any]:
        for func in functions:
            data = func(data)
        return data

    return apply_all


def modify_components(
    *transform_functions: Callable[[dict[str, Any]], dict[str, Any]],
) -> Callable[[dict[str, Any]], dict[str, Any]]:
    """Apply multiple transformations to components."""
    return compose(*transform_functions)


def apply_property_map(component: dict[str, Any], property_map: dict[str, str]) -> dict[str, Any]:
    """Apply a key mapping to component keys.

    For each key in the `component` dictionary, if the key exists in the `property_map`,
    it will be replaced by the corresponding mapped key. Otherwise, the original key is used.

    Parameters
    ----------
    component : dict[str, Any]
        The original dictionary where keys represent component properties.
    property_map : dict[str, str]
        A dictionary mapping old property names to new property names.

    Returns
    -------
    dict[str, Any]
        A new dictionary where the keys have been remapped according to `property_map`.

    Examples
    --------
    >>> component = {"voltage": 230, "current": 10}
    >>> property_map = {"voltage": "v", "current": "i"}
    >>> apply_property_map(component, property_map)
    {'v': 230, 'i': 10}

    >>> component = {"voltage": 230, "resistance": 50}
    >>> property_map = {"voltage": "v", "current": "i"}
    >>> apply_property_map(component, property_map)
    {'v': 230, 'resistance': 50}
    """
    return {property_map.get(key, key): value for key, value in component.items()}


def apply_pint_deconstruction(component: dict[str, Any], unit_map: dict[str, str]) -> dict[str, Any]:
    """Get property magnitude from a pint Quantity.

    If unit_map is passed, convert to units specified

    Parameters
    ----------
    component: dict[str, Any]
        Dictionary representation of component. Typically created with `.model_dump()`
    unit_map: dict[str, str]
        Map to convert property to a desired units. Optional.
    """
    return {
        property: get_property_magnitude(property_value, to_unit=unit_map.get(property, None))
        for property, property_value in component.items()
    }


def apply_valid_properties(
    component: dict[str, Any], valid_properties: list[str], add_name: bool = False
) -> dict[str, Any]:
    """Filter a component dictionary to only include keys that are in the valid properties list.

    Parameters
    ----------
    component : dict of str to Any
        A dictionary representing the component with properties as keys.
    valid_properties : list of str
        A list of valid property names. Only keys present in this list will be kept in the output.

    Returns
    -------
    dict
        A new dictionary with only the properties from `valid_properties`.

    Examples
    --------
    >>> component = {"voltage": 230, "current": 10, "resistance": 50}
    >>> valid_properties = ["voltage", "current"]
    >>> apply_valid_properties(component, valid_properties)
    {'voltage': 230, 'current': 10}
    """
    if add_name:
        default_properties = ["name"]
        return {
            key: value
            for key, value in component.items()
            if key in valid_properties or key in default_properties
        }
    return {key: value for key, value in component.items() if key in valid_properties}


def apply_unnest_key(component: dict[str, Any], key_map: dict[str, Any]) -> dict[str, Any]:
    """Unnest specific nested dictionary values based on a provided key map.

    This function processes a dictionary, potentially containing nested dictionaries,
    and unnests specific values based on a provided key map. For each key in the input
    dictionary that has a corresponding entry in the key map, if the value is a dictionary,
    it extracts the value using the mapped key.

    Parameters
    ----------
    component : dict[str, Any]
        The input dictionary to process.
    key_map : dict[str, Any]
        A dictionary mapping keys in the input dictionary to keys in nested dictionaries.

    Returns
    -------
    dict[str, Any]
        A new dictionary with unnested values based on the key map.

    Examples
    --------
    >>> component = {
    ...     "name": "Example",
    ...     "config": {"type": "A", "value": 10},
    ...     "data": {"content": "Some data"},
    ... }
    >>> key_map = {"config": "type", "data": "content"}
    >>> apply_unnest_key(component, key_map)
    {'name': 'Example', 'config': 'A', 'data': 'Some data'}

    Notes
    -----
    - If a key in the input dictionary is not in the key map, its value remains unchanged.
    - If a key is in the key map but the corresponding value in the input dictionary
      is not a dictionary, the value remains unchanged.
    - If a key is in the key map and the corresponding value is a dictionary, but the
      mapped key is not in this nested dictionary, the result for this key will be None.
    """
    if not key_map:
        return component
    return {
        key: value if not isinstance(value, dict) else value.get(key_map.get(key), value)
        for key, value in component.items()
    }


def apply_default_value(component: dict[str, Any], default_value_map: dict[str, Any]) -> dict[str, Any]:
    """Unnest specific nested dictionary values based on a provided key map.

    This function iterates over a dictionary (`component`) and applies default values
    to its missing or `None` values using a `default_value_map`. For each key in
    `default_value_map`, if the key is not present in `component` or has a `None` value,
    it will be assigned the corresponding default value from `default_value_map`.

    Parameters
    ----------
    component : dict[str, Any]
        The input dictionary to process.
    default_value_map : dict[str, Any]
        A dictionary specifying default values. Each key in this dictionary corresponds
        to a key in `component`, and the associated value is the default value to apply
        if the key is missing or `None` in `component`.

    Returns
    -------
    dict[str, Any]
        A new dictionary with unnested values based on the key map.

    Examples
    --------
    >>> component = {"name": "example", "year": None}
    >>> default_value_map = {"year": 2024, "month": "M10"}
    >>> apply_default_value(component, default_value_map)
    {'name': 'example', 'year': 2024, 'month': 'M10'}

    >>> component = {"name": "example", "year": 2023}
    >>> default_value_map = {"year": 2024, "month": "M10"}
    >>> apply_default_value(component, default_value_map)
    {'name': 'example', 'year': 2023, 'month': 'M10'}
    """
    if not default_value_map:
        return component
    for key in default_value_map:
        if key not in component:
            component[key] = default_value_map[key]
        if component.get(key, None) is None:
            component[key] = default_value_map[key]
    return component


def get_property_magnitude(property_value, to_unit: str | None = None) -> Any:
    """Return magnitude with the given units for a pint Quantity.

    Parameters
    ----------
    property_name

    property_value
        pint.Quantity to extract magnitude from
    to_unit
        String that contains the unit conversion desired. Unit must be compatible.

    Returns
    -------
    float
        Magnitude representation of the `pint.Quantity` or original value.
    """
    if not isinstance(property_value, pint.Quantity | BaseQuantity):
        return property_value
    if to_unit:
        unit = to_unit.replace("$", "usd")  # Dollars are named usd on pint
        property_value = property_value.to(unit)
    return property_value.magnitude


def apply_flatten_key(d: dict[str, Any], keys_to_flatten: set[str]) -> dict[str, Any]:
    """Flatten the specified keys in a dictionary by merging their sub-keys into the main dictionary
    with the key's name prefixed to each sub-key.

    Parameters
    ----------
    d : dict
        The input dictionary, where some values are dictionaries to be flattened.
    keys_to_flatten : list of str
        The keys in the dictionary `d` to be flattened if they contain sub-keys.

    Returns
    -------
    dict
        A new dictionary with the selected keys flattened. Other keys remain unchanged.

    Examples
    --------
    >>> d = {"x": {"min": 1, "max": 2}, "y": {"min": 5, "max": 10}, "z": 42}
    >>> flatten_selected_keys(d, ["x"])
    {'x_min': 1, 'x_max': 2, 'y': {'min': 5, 'max': 10}, 'z': 42}

    >>> flatten_selected_keys(d, ["y"])
    {'x': {'min': 1, 'max': 2}, 'y_min': 5, 'y_max': 10, 'z': 42}
    """
    flattened_dict = {}

    for key, value in d.items():
        if key in keys_to_flatten and isinstance(value, dict):
            for inner_key, inner_value in value.items():
                flattened_dict[f"{key}_{inner_key}"] = inner_value
        else:
            flattened_dict[key] = value

    return flattened_dict


def apply_extract_key(d: dict[str, Any], key: str, keys_to_extract: set[str]) -> dict[str, Any]:
    """Extract keys from a nested dictionary and put them in the first level if specific conditions are met.

    Parameters
    ----------
    d : dict
        The input dictionary that may contain nested dictionaries.
    key : str
        The key in the input dictionary whose value should be a nested dictionary.
    keys_to_extract : set[str]
        The set of keys to extract from the nested dictionary.

    Returns
    -------
    dict
        If conditions are met: a new dictionary containing all original key-value pairs
        plus the extracted key-value pairs at the top level.
        If conditions are not met: returns the input dictionary unchanged.

    Notes
    -----
    The function will return the input dictionary unchanged if any of these conditions are met:
    - The specified key is not in the input dictionary
    - Any of keys_to_extract already exist in the top level of input dictionary
    - None of keys_to_extract exist in the nested dictionary

    Examples
    --------
    >>> component = {"name": "Gen01", "rating": {"up": 100, "down": -100}}
    >>> apply_extract_key(component, "rating", {"up"})
    {'name': 'Gen01', 'rating': {'up': 100, 'down': -100}, 'up': 100}

    >>> # Returns unchanged when extracted key already exists at top level
    >>> d = {"name": "Gen01", "rating": {"up": 100}, "up": 50}
    >>> apply_extract_key(d, "rating", {"up"})
    {'name': 'Gen01', 'rating': {'up': 100}, 'up': 50}

    >>> # Returns unchanged when no keys_to_extract exist in nested dict
    >>> apply_extract_key(component, "rating", {"middle"})
    {'name': 'Gen01', 'rating': {'up': 100, 'down': -100}}
    """
    if key not in d.keys() or any(k in d.keys() for k in keys_to_extract):
        return d

    if not any(k in d[key].keys() for k in keys_to_extract):
        return d

    extracted_keys = {key: value for key, value in d[key].items() if key in keys_to_extract}
    return {**d, **extracted_keys}



"""Model related to branches."""

from typing import Annotated

from infrasys.value_curves import InputOutputCurve
from pydantic import Field, NonNegativeFloat, NonPositiveFloat

from r2x.models.core import Device, FromTo_ToFrom, MinMax
from r2x.models.topology import ACBus, Arc, Area, DCBus
from r2x.units import ActivePower, Percentage


class Branch(Device):
    """Class representing a connection between components."""

    @classmethod
    def example(cls) -> "Branch":
        return Branch(name="ExampleBranch")


class ACBranch(Branch):
    """Class representing an AC connection between components."""

    arc: Annotated[Arc | None, Field(description="The branch's connections.")] = None
    from_bus: Annotated[ACBus, Field(description="Bus connected upstream from the arc.")]
    to_bus: Annotated[ACBus, Field(description="Bus connected downstream from the arc.")]
    r: Annotated[float | None, Field(description=("Resistance of the branch"))] = None
    x: Annotated[float | None, Field(description=("Reactance of the branch"))] = None
    rating: Annotated[ActivePower, Field(ge=0, description="Thermal rating of the line.")] | None = None


class MonitoredLine(ACBranch):
    """Class representing an AC transmission line."""

    b: Annotated[FromTo_ToFrom | None, Field(description="Shunt susceptance in pu")] = None
    g: Annotated[FromTo_ToFrom | None, Field(description="Shunt conductance in pu")] = None
    rating_up: Annotated[ActivePower, Field(ge=0, description="Forward rating of the line.")] | None = None
    rating_down: Annotated[ActivePower, Field(le=0, description="Reverse rating of the line.")] | None = None
    losses: Annotated[Percentage, Field(description="Power losses on the line.")] | None = None

    @classmethod
    def example(cls) -> "MonitoredLine":
        return MonitoredLine(
            name="ExampleMonitoredLine",
            from_bus=ACBus.example(),
            to_bus=ACBus.example(),
            losses=Percentage(10, "%"),
            rating_up=ActivePower(100, "MW"),
            rating_down=ActivePower(-100, "MW"),
            rating=ActivePower(100, "MW"),
        )


class Line(ACBranch):
    """Class representing an AC transmission line."""

    b: Annotated[FromTo_ToFrom | None, Field(description="Shunt susceptance in pu")] = None
    g: Annotated[FromTo_ToFrom | None, Field(description="Shunt conductance in pu")] = None
    active_power_flow: NonNegativeFloat
    reactive_power_flow: NonNegativeFloat
    angle_limits: MinMax

    @classmethod
    def example(cls) -> "Line":
        return Line(
            name="ExampleLine",
            from_bus=ACBus.example(),
            to_bus=ACBus.example(),
            rating=ActivePower(100, "MW"),
            active_power_flow=100,
            reactive_power_flow=100,
            angle_limits=MinMax(min=-0.03, max=0.03),
        )


class Transformer2W(ACBranch):
    """Class representing a 2-W transformer."""

    active_power_flow: NonNegativeFloat = 0.0
    reactive_power_flow: NonNegativeFloat = 0.0
    primary_shunt: float | None = None

    @classmethod
    def example(cls) -> "Transformer2W":
        return Transformer2W(
            name="Example2WTransformer",
            rating=ActivePower(100, "MW"),
            from_bus=ACBus.example(),
            to_bus=ACBus.example(),
            active_power_flow=100,
            reactive_power_flow=100,
        )


class TapTransformer(ACBranch):
    active_power_flow: NonNegativeFloat
    reactive_power_flow: NonNegativeFloat
    primary_shunt: float | None = None
    tap: Annotated[
        NonNegativeFloat,
        Field(
            ge=0,
            le=2.0,
            description=(
                "Normalized tap changer position for voltage control, varying between 0 and 2, with 1"
                "centered at the nominal voltage"
            ),
        ),
    ]


class PhaseShiftingTransformer(ACBranch):
    active_power_flow: NonNegativeFloat
    reactive_power_flow: NonNegativeFloat
    primary_shunt: float | None = None
    tap: Annotated[
        NonNegativeFloat,
        Field(
            ge=0,
            le=2.0,
            description=(
                "Normalized tap changer position for voltage control, varying between 0 and 2, with 1"
                "centered at the nominal voltage"
            ),
        ),
    ]
    α: Annotated[float, Field(ge=-1.571, le=1.571)]
    # primary_shunt: float | None = None
    phase_angle_limits: MinMax


class DCBranch(Branch):
    """Class representing a DC connection between components."""

    from_bus: Annotated[DCBus, Field(description="Bus connected upstream from the arc.")]
    to_bus: Annotated[DCBus, Field(description="Bus connected downstream from the arc.")]


class AreaInterchange(Branch):
    """Collection of branches that make up an interfece or corridor for the transfer of power."""

    active_power_flow: NonNegativeFloat
    flow_limits: FromTo_ToFrom
    from_area: Annotated[Area, Field(description="Area containing the bus.")] | None = None
    to_area: Annotated[Area, Field(description="Area containing the bus.")] | None = None

    @classmethod
    def example(cls) -> "AreaInterchange":
        return AreaInterchange(
            name="ExampleAreaInterchange",
            active_power_flow=10.0,
            flow_limits=FromTo_ToFrom(from_to=100, to_from=-100),
            from_area=Area.example(),
            to_area=Area.example(),
        )


class TModelHVDCLine(DCBranch):
    """Class representing a DC transmission line."""

    rating_up: Annotated[NonNegativeFloat, Field(description="Forward rating of the line.")] | None = None
    rating_down: Annotated[NonPositiveFloat, Field(description="Reverse rating of the line.")] | None = None
    losses: Annotated[NonNegativeFloat, Field(description="Power losses on the line.")] = 0
    resistance: Annotated[NonNegativeFloat, Field(description="Resistance of the line in p.u.")] | None = 0
    inductance: Annotated[NonNegativeFloat, Field(description="Inductance of the line in p.u.")] | None = 0
    capacitance: Annotated[NonNegativeFloat, Field(description="Capacitance of the line in p.u.")] | None = 0

    @classmethod
    def example(cls) -> "TModelHVDCLine":
        return TModelHVDCLine(
            name="ExampleDCLine",
            from_bus=DCBus.example(),
            to_bus=DCBus.example(),
            rating_up=100,
            rating_down=80,
        )


class TwoTerminalHVDCLine(ACBranch):
    """Class representing a DC transmission line."""

    active_power_flow: NonNegativeFloat
    active_power_limits_from: MinMax
    active_power_limits_to: MinMax
    reactive_power_limits_from: MinMax
    reactive_power_limits_to: MinMax
    loss: InputOutputCurve
